/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

//  xxxxxxxxxxxxxxx           version: 1.0           xxxxxxxxxxxxxxxx     

//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx


// These variables below can be directly modified by the user on the options form: 
// Commented variables are not globally (window scope) defined 

var minute = 1;       // Task time in minutes

var raster = 10;       //  The size of the raster.  canvas (viewport) raster size

// var gtype = '2';	    // Stimulus generator type
 	
// var tcolor= '3';   //  color type : "1" => Full random color;  "2" => Only basic colors;  "3" =>Only black color
	   
// var demo = false;     // if (demo == true) =>  In demo mode, each subsequent correct selection will be different color

// var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 


// Other global constants and variables:

var timer = 60;     // Task time in seconds. see stimGen() 

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var cnt = 0;   					// global timer cnt (incresing by clock frequency 10ms)

var sec = 0;            // global timer counter for seconds

var is_task = false;    // global flag: task in progress or not. see appTimer(). is_task == true for all events between startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 
 
var is_storage= false; // local storage is enabled on user browser or not

var startTime;    // start time for task

var stim = [];   //  stimulus array. see stimGen() and drawStim() functions

var cells = [];  // array for each cell. cell is currently selected => 1. cell is currently not selected => 0

var res_cel= [];  // time point (iteration) when all cells have to be unselected. see setTargets() 

/* dependent variable: correct choices. if gtype == '1' index: 
   [ 0 => total targets; 1 => M-type targets correct detected; 2 => P-type targets correct detected ] 
   else index: 
   [ 0 => total targets; 1 => M-type targets correct detected; 2 => P-type targets correct detected; 3 => left/up targets correct detected; 4 => right/down targets correct detected] */ 

var cor = [];    

var errors;   // dependent variable: numbers of errors in task (simple counter)


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [create/start Timer, check storage, show start panel..] 
function init() 
{    
  
	if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},10); // each 10 ms! 
  
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }		

  initRaster();
	
	setIns();

	showStart();	
}


// try to set raster to "optimal" value
function initRaster()
{
	let rst= 20, w, h, r;
	for (let j = 2; j <= 20; j+= 2)  
	{
		w = window.innerWidth / j;  // cell width,
		h= window.innerHeight / j;  // cell height
		r= Math.round(0.9 * Math.min(w,h)/2); // P-cells 
		if ( r < 50 )	 { 
   		rst= j;
			break;
	  }
	}	
	w= window.innerWidth / rst;
	h= window.innerHeight / rst;  // cell width, cell height
	r= Math.round(0.9 * Math.min(w,h)/2); // P-cells 
	document.getElementById('raster_msg').textContent= 'The total number of stimuli displayed on the screen= ' + ( rst * rst ) + '. Shape resolution= ' + r + ' pixels.'; //Recommended minimum= 50 pixels.'; 
	
	document.getElementById('raster').value= (rst % 2 == 0 && rst >= 2 && rst <= 20) ? rst : 2;
	
	if (document.getElementById('raster').validity.valid) {
		raster= parseInt(document.getElementById('raster').value, 10); 
  }
}


// application timer handler function 
function appTimer()  
{    
	if (is_task) // else: nop, idle, do nothing 
	{  
		if ( sec > timer-1 )  {     // task time is limited
			is_task= false;  
			theEnd();
		}	
		else  {
			if ( cnt % 100 == 0 && res_cel[sec] && res_cel[sec] == 1 ) {
				initCells();
			}
			drawStim(stim[sec], false);			
		}
		cnt++;  // each 10ms
		if ( cnt % 100 == 0 ) {
			sec++;   // add one "second"
		}
	}  
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if (is_task) 
	{ 
		let c= document.getElementById('canvas');
		c.width= window.innerWidth;  
		c.style.width= canvas.width + 'px';	     
		c.height= window.innerHeight;   
		c.style.height= canvas.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 		 		
		
		drawStim(stim[sec],true);	  // repaint stimulus		
	}
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// set instructions text depending on user selected task options
function setIns()  
{ 
  
	document.getElementById('instructions').style.display= 'block';		       
    
	document.getElementById('ins_tcolor').style.display= document.getElementById('tcolor').value == '3' ? 'none' : 'block';	
		
	document.getElementById('ins_minute').textContent= minute;
	
	document.getElementById('ins_demo').style.display= document.getElementById('demo').checked ? 'block' : 'none';	
  document.getElementById('ins_demo2').style.display= (document.getElementById('demo').checked && document.getElementById('tcolor').value != '3') ? 'block' : 'none';	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty= true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
			
		case 'minute':		
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value*60);  
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			}			
			break;  			
			
		case 'raster':		
			if (e.validity.valid)  {
				let rst= parseInt(e.value,10);
				let w =  window.innerWidth / rst, h= window.innerHeight / rst;  // cell width, cell height
	      let r= Math.round(0.9 * Math.min(w,h)/2); // P-cells 
				msg.textContent= 'The total number of stimuli displayed on the screen= ' + ( rst * rst ) + '. Shape resolution= ' + r + ' pixels.'; //Recommended minimum= 50 pixels.';  
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			}			
			break; 
			
    case 'gtype':
      switch (e.value)  
	    {
				case '1':	
				  msg.textContent= 'Spatial location of target stimuli is random generated. Results at the end of the task ignore the spatial location of the target stimuli.';
				  break;
				case '2':	
				  msg.textContent= '50% of the target stimuli will be in the left half of the screen. The other 50% of the target stimuli will be in the right half of the screen.';
				  break;
				case '3':	
				  msg.textContent= '50% of the target stimuli will be in the upper half of the screen. The other 50% of the target stimuli will be in the lower half of the screen.';
				  //break;
			}
			break;	
		
    case 'tcolor':                                                                                                                                            
      if (e.value == '1') {
        msg.textContent= 'All stimuli will be displayed in all possible colors (when possible). Caution: If a color similar to white is generated, you will not be able to distinguish the drawing from the background color.';
      }
      else if (e.value == '2') {
        msg.textContent= 'All stimuli will be displayed only in basic html colors (when possible)';
      } 
      else  {
        msg.textContent= 'All stimuli will be drawn in black only';
      }      
      break;

    case 'demo': 
			msg.textContent= e.checked ? 'On. (all target stimuli will have this background color)' : 'Off';    
			break;

    case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty == true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number". see: setPar()
function optionsValid()  
{ 
  let res= document.getElementById('minute').validity.valid && document.getElementById('raster').validity.valid;
  return res;              
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// start task (research) 
function startTask(e)   
{     
  	
  initLayout();     			
  
	setPar();
	
  stimGen();
	
	is_task= true;    // finally "activate" timer   
	      
}


// init visible/invisible html dom elements, progress bar and background colors
function initLayout()
{   	 
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';
 
  let c= document.getElementById('canvas');
	c.style.display= 'block';
	c.width= window.innerWidth;  
	c.style.width= canvas.width + 'px';	     
	c.height= window.innerHeight;   
	c.style.height= canvas.height + 'px';  	
}	


// set internal application research parameters (global vars)
// reset (init) global (window) variables
function setPar()  
{    	  
	stim= [];
	errors= 0;		
	
		// this 2 lines bellow are probably redundant (danger of infinite loop)
		if ( raster % 2 != 0 ) { raster--; }
		if ( raster < 2  ) { raster= 2; }
	
	if ( document.getElementById('gtype').value == '1' )  {
		cor = [0, 0, 0];  // index: [ 0 => total targets; 1 => M-type targets correct detected; 2 => P-type targets correct detected ]  
	}
	else  {
    cor = [0, 0, 0, 0, 0]; // index: [ 0 => total targets; 1 => M-type targets correct detected; 2 => P-type targets correct detected; 3 => left/up targets correct detected; 4 => right/down targets correct detected] 
	}	
	
	res_cel= [];
	
	initCells();
  
	cnt= 0;       // reset timer counters (each 10ms)
	sec= 0;  	    // timer counters (seconds)

}


// init all raster cells as unselected
function initCells() {
	const N = raster * raster;   // N = number of cells (stimulus count) in canvas object
	cells= [];
	for (let j = 0; j < N; j++)  {
		cells[j]= 0;   // not selected
	}
}


/* return total number of targets in one "minute"
   set n == total number of target stimuli in minute (m + p targets)
   return range of 6..18 target stim per minute  */
function targetPerMin()  {
	let n;  // function result have this condition: n % 2 == 0
	let rnd= getRndInt(1,6);
	switch (rnd)  
	{  
	  case 1: // equal time intervals 
			n= 6; // dt = 10 ( 6 * 10 = 60 )
	    break;
			
	  case 2: // unequal time intervals 
			n= 8; // dt = Math.round(60/8) = 8 ( 7 * 8 + 4 = 60 )
			break;
			
		case 3:   // equal time intervals 
			n= 10; // dt= 6 ( 10 * 6 = 60 )
			break;
			
    case 4:	 // equal time intervals 	
			n= 12; // dt = 5 ( 12 *5 = 60 )
			break;
		
		case 5:	 // unequal time intervals		
			n= 14; // dt = Math.round(60/14) = 4 ( 13 * 4 + 8 = 60)
			break;
		
		default:  // unequal time intervals
      n= 18;  // dt = Math.round(60/18) = 3 ( 17 * 3 + 9 = 60 )  
      //break;			
	}
	return n;
}	


/* this function determines: in which second within each minute the target stimuli will appear
   and at what time point ("second") all cells must be reset 
	 n = total number of target stimuli in minute  
	 off = time offset for res_cel array (array offset) */
function setTarget(n,off)
{ 
	let res= [];
	const dt= Math.round(60/n); // time interval between target stimuli  
	let tot= 0;
	res_cell= [];
	while ( tot < n )  { 				
		res[tot*dt]= 1;  // at this time point ("second") it will be presented target stimulus
		res_cel[tot*dt+off]= 1;   // at this time point or iteration all cells must be reset 
		tot++;    		
	} 
	return res;
}	


/* stype = Stimulus generator type [ '1' => 50% M  + 50% P ; '2' => 50% left + 50% right ; '3' => 50% up + 50% down ] = document.getElementById('gtype').value 
   rst = window.raster => canvas cells raster
	 i_ex, j_ex = cells that contained previous target stimuli (must not be repeated) 
	 return array of positions of target cells in raster  */
function getTargetCells(stype, rst, i_ex, j_ex)
{
	const N = rst * rst -1;   // N = number of cells (stimulus count) in canvas object
	let i, j, row, col, res= [];
	
	if ( stype == '1' )  
	{  // random
		do  {
			i= getRndInt(0,N);  // i == M - target cell
			j= getRndInt(0,N);  // j == P - target cell
		} while ( i == j || i == i_ex || i == j_ex  || j == i_ex || j == j_ex);
	}
	
	else if ( stype == '2' )
	{ // left - right 
		do  {
			i= getRndInt(0,N);  // i => index 0 == left  
		  col = Math.floor (i % rst);  
		} while ( col >= rst / 2 || i == i_ex || i == j_ex );
		do  {
			j= getRndInt(0,N);  // j => index 1 == right   
		  col = Math.floor (j % rst);   
		} while ( col < rst / 2 || j == i_ex || j == j_ex );
	}
	
	else if ( stype == '3' )
	{ // up - down 
		do  {
			i= getRndInt(0,N);  // i => index 0 == up 
		  row = Math.floor (i / rst);  
		} while ( row >= rst / 2 || i == i_ex || i == j_ex );
		do  {
			j= getRndInt(0,N);  // j => index 1 == down 
		  row = Math.floor (j / rst);   
		} while ( row < rst / 2 || j == i_ex || j == j_ex );
	}	
	
	res[0]= i;  // 1. target cell
	res[1]= j;  // 2. target cell
	
	return res;
}


// stimulus generator
function stimGen()
{
	const rst = raster; // speed? 
	const N = rst * rst;   // N = number of cells (stimulus count) in canvas object
	const m_tar= -1;  // target stimulus for M-type
	const p_tar= 9;   // target stimulus for P-type
	const ball = 10;  // ball (circle) radius (px) => M-type stimulus

	let tcolor= document.getElementById('tcolor').value;	
  let gtype= document.getElementById('gtype').value;
	let sec;  // "seconds" within each "minute". range: 0..59	
	let n;    // number of targets in current minute
  let	target;  // "seconds" in each target start (for current minute)
  let	dt;      // time interval between target stimuli
  let tcells;   // position of target cells. index [0] = M-target, [1] = P-target 	
	let off = 0;  // parameter for setTarget(). it points when all cells have to be reset (selected or not) 
	let prev_1= null, prev_2= null;  // spatial positions of the cell for the previous target position 
	let first, second;  // first and second target parallel stimulus 
		
	timer = minute * 60;    //   update timer (total number of "seconds") = number of minute * 60
	
	for (let i= 0; i< timer; i++)  // i = "second" (time cycle)
	{ 	
    stim[i]= [];  // open new jagged array for this second
		
		if ( i == 0 || sec > 59 )  {
      sec= 0;  // reset "seconds" counter within each "minute". range: 0..59	
			n= targetPerMin();     // number of targets in current minute
			cor[0]+= 2*n;     // increase total number of target stimulus
			target= setTarget(n, off);  // "seconds" for each target start (for current minute)
			dt= Math.round(60/n);  // time interval between target stimuli. see setTarget() function
			off+=60;  			
		}						  
		
		if ( target[sec] && target[sec] == 1 )			// begin of new target stimulus
		{ 
			tcells= getTargetCells(gtype, rst, prev_1, prev_2);
      prev_1= tcells[0];  // save target cells positions for next iteration
      prev_2= tcells[1];
			if ( Math.random() < 0.5 )  {
				first=  m_tar;    // set target stimulus for M-type or/and P-type 
				second= p_tar;    
			}
			else  {
				first=  p_tar;    // rotate 1. and 2. target stimulus position (M-type and P-type stimulus)
				second= m_tar;   
			}				
		}	
		
		sec++;		// add new "second"
					
		for (let j=0; j< N; j++)  // => j = cell (shape)
		{						
			
			let t;  // shape type
			if ( j == tcells[0] || j == tcells[1] ) {
				t = (j == tcells[0]) ? first : second;   // set target stimulus for M-type or/and P-type        
			}
			else  
			{				
				do {
					t = getRndInt(-36,36);  // shape type
				} while ( t == m_tar || t == p_tar ); // shape type must be different from target stimulus (M and P type) 
		  }
		
			let col= newColor(tcolor);    // get random color based on: document.getElementById('tcolor').value  			  
			if ( t >= 0 )  {
			  stim[i].push([t, col]); 	// // P-type stim = [0 = stim type, 1 = color]
			}
			else  
			{ 			
				// M-type stim = [0 = stim type, 1 = color, 2..N = shape type specific]						
				switch ( t )
				{					
					case -1:	 // horizontal movement
					case -2:   // vertical movement
						// stim = [0 = stim type, 1 = color,  2 = x, 3 = y, 4 = dx or dy]	
						stim[i].push([t, col, null, null, ball]);  
						break;  
					
					case -3:    // rotation around cell center => clockwise
					case -4:		// rotation around cell center => counterclockwise						
						// stim = [0 = stim type, 1 = color, 2 = polar angle] 
						stim[i].push([t, col, getRndFloat(0, 2*Math.PI)]);  
						break;

					case -5:		// rotation from cell center to cell edge => clockwise		
					case -6:		// rotation from cell center to cell edge => counterclockwise						
					  // stim = [0 = stim type, 1 = color, 2= polar angle; 3= current radius, 4= radius increase (decrease) in next call]										
						stim[i].push( [t, col, 0, getRndFloat(0, 2*Math.PI), ball, 1] );
						break;  							
					
					case -7:	 // movement that mimics Archimedean spiral from cell center to cell edge and back => start in cell center with clockwise	rotation
					case -8:   // movement that mimics Archimedean spiral from cell center to cell edge and back => start in cell center with counterclockwise rotation                    					
						// stim = [0 = stim type, 1 = color, 2= polar angle; 3= current radius, 4= radius increase (decrease) in next call]												
						if ( t == - 8)  {
							stim[i].push( [ t, col, getRndFloat(0, 2*Math.PI), ball, -1 ] );  
						}
						else  {					 						
							stim[i].push( [ t, col, getRndFloat(0, 2*Math.PI), ball,  1 ] );  
						}
						break;					
					
					case -9:  // random ball moving within cell walls => clockwise
						// stim = [0 = stim type, 1 = color, 2 = x, 3 = y, 4 = dx, 5 = dy]	
						stim[i].push([t, col, null, null, ball, 0 ]);						
						break;
					
					case -10:  // random ball moving within cell walls => counterclockwise						
						// stim = [0 = stim type, 1 = color, 2 = x, 3 = y, 4 = dx, 5 = dy]	
						stim[i].push([t, col, null, null, -ball, 0 ]);					
						break;
						
					default:  // random ball moving within cell walls            
					  // stim = [0 = stim type, 1 = color,  2 = x, 3 = y, 4 = dx, 5 = dy]						
						stim[i].push([t, col, null, null, ball, ball ]);
						//break;
				}  // switch ( t )
				
			}  // else if ( t >= 0 )  
					
		} // for (let j=0; j<N; j++)  // => j = cell (shape)
			
	}		// for (let i= 0; i< timer; i++)  // i = second (cycle)
	
	startTime = new Date();	
}	


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


/* https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Math/random
Returns a float random number between the specified values. 
The returned value is no lower than (and may possibly equal) min, and is less than (and not equal) max.
Returns a random number between min (inclusive) and max (exclusive)  */
function getRndFloat(min, max) {
	return Math.random() * (max - min) + min;
}  	


// get new random color based on window.tcolor
// par tcolor => color type : "1" => Full random color;  "2" => Only basic colors;  "3" =>Only black color
function newColor(tcolor)
{
  if (tcolor == '1')  {
    return getRndColor();
  }
  else if (tcolor == '2')  {
    return getColor();   
  }
  else  {
    return '#000';  
  }
}


// sets one of mostly basic html predefined colors
// this is mostly default set of 16 html Basic colors 
function getColor() 
{
  let clr = [    
    '#000000',     /* "black" */   
    '#FF0000',     /* "red"   */
    '#008000',     /* "green" */
    '#0000FF',     /* "blue" */   
    '#FFFF00',     /* "yellow" */     
    '#00FF00',     /* "lime" */     
    '#00FFFF',     /* "cyan" / "aqua" */     
    '#FF00FF',     /* magenta / fuchsia  */  
    '#C0C0C0',     /* "silver" */      
    '#800000',     /* maroon */     
    '#808000',     /* "olive" */  
    '#800080',     /* "purple" */    
    '#008080',     /* teal */      
    '#000080',     /* navy */    
    '#FFC0CB',     /* "pink" */           
    '#FFA500'      /* "orange" */ 
    ];      
  let rnd= getRndInt(0,15);  // array clr limits
  //var rnd=  Math.floor(Math.random()*16);   // 0..15
  return clr[rnd];
}


// get full random colors in hex format
function getRndColor() 
{
  let letters = '0123456789abcdef';
  let color = '#';
  for (let i=0; i<6; i++) {
    color += letters[Math.floor(Math.random() * 16)];     // returns a random integer from 0 to 15 
  }
  return color;
}  


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	if (document.getElementById('minute').validity.valid) {
		minute= Math.round(parseInt(document.getElementById('minute').value, 10)); 
  }
	
	if (document.getElementById('raster').validity.valid) {
		raster= Math.round(parseInt(document.getElementById('raster').value, 10));  
  }
}


/* draw (repaint) all stimuli         
  parameters: st => jagged stimulus array instance for current cycle ("second") 
  is_res => reset startup coordinates (for moving shapes - M type) if window is resized  */
function drawStim(st,is_res) 
{  
 	const m_tar= -1;  // target stimulus for M-type
	const p_tar= 9;   // target stimulus for P-type	
  const is_demo= document.getElementById('demo').checked;
	const rst = raster; //  rst = window.raster 
	let c= document.getElementById('canvas'), ctx= c.getContext('2d');  
	ctx.clearRect(0, 0, c.clientWidth, c.clientHeight);  // clear canvas object
			
	const ball= 10; // ball radius (px)		
	const width = c.width / rst, height = c.height / rst;  // cell width, cell height
	const rp= Math.round(0.9 * Math.min(width,height)/2); // radius for P-cells 			 	
	const len = rst * rst;   // total number of cells
	
	for (let j= 0; j < len; j++)
	{	   
	
		let row = Math.floor (j / rst);
		let col = Math.floor (j % rst);                 						
		let left= col * width, top = row * height;  // cell left bound, cell top bound
		let x, y, dx, dy, ang, max_r;
		let r= Math.min(width,height)/2;  // M-cells default radius (case -3 .. case -8  override this r)
		
		ctx.beginPath();     // drawGrid() => this section draw square border for each shape => rectangle 
    ctx.rect(left, top, width, height);   //  	context.rect(x,y,width,height);
    		
		let stype= st[j][0];   // stimulus type 		
		if ( (cells[j] == 1) || ( is_demo && ( stype == m_tar || stype == p_tar ) ) )     
    { // set background color for selected answer or set background color in demo mode for correct answer. you must call before drawShape() handler               
      // Gray90 == selected or Ivory == correct 
      ctx.fillStyle=  cells[j] == 1  ?  '#e5e5e5' : '#fffff0';  // the selected answer takes precedence over the correct answer (background color)  
			ctx.strokeStyle= '#f00';
    }
    else  {
      ctx.fillStyle= '#fff';
      ctx.strokeStyle= '#000';
    }  
    ctx.fill();             
    ctx.stroke();											
		if  ( stype >=  0 )  {  //  stim[i].push( [stimulus type, stimulus color] );
      drawShape(ctx, left + width/2, top + height/2, rp, stype, st[j][1]); // function drawShape(ctx, x, y, r, shape, col)			
    }  
		else 
		{
			ctx.fillStyle= st[j][1];  // load (update) M-type stimulus color 
			switch ( stype )
			{				
				// M-type stim = [0 = stim type, 1 = color, 2..N = shape type specific]	
				
				case -1:    // horizontal movement				
					// stim = [0 = stim type, 1 = color,  2 = x, 3 = y, 4 = dx or dy]  
					x= st[j][2], y= st[j][3], dx= st[j][4];
					if (y == null || is_res)  {		 // init ball x,y, start coord somewhere in shape rect (cell) 					 
						x = getRndInt( left + ball, Math.floor(left + width - ball) );
						y = top + height/2;  
						st[j][3]= y;           		
					}
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();   
					// check and update dx in next call 
					if ( x + dx < left + ball || x + dx > left + width - ball ) {
						dx = -dx;											
					}
					st[j][2]= x+dx;
					st[j][4]= dx;				
					break;
			
				case -2:    // vertical movement
					// stim = [0 = stim type, 1 = color,  2 = x, 3 = y, 4 = dx or dy]				
					x= st[j][2], y= st[j][3],  dy= st[j][4];
					if (x == null || is_res)  {		 // init ball x,y, start coord somewhere in shape rect (cell) 					 
						x = left + width/2;
						y = getRndInt( top + ball, Math.floor(top + height - ball) ); 
						st[j][2]= x;          		
					}
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();   
					// check and update dy in next call 
					if ( y + dy < top + ball || y + dy > top + height - ball ) {
						dy = -dy;														          
					}
					st[j][3]= y+dy;
					st[j][4]= dy;
					break;
					
				case -3:    // rotation around cell center => clockwise
				case -4:		// rotation around cell center => counterclockwise			
					// stim = [0 = stim type, 1 = color, 2 = polar angle] 			
					ang= st[j][2];  // units: radians. must always be init ok
					r = Math.min(width,height)/2 - ball;
					x = left + width/2 + r * Math.cos(ang);
					y = top + height/2 + r * Math.sin(ang);
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();   														          
					st[j][2]+= st[j][0] == -3 ? 2*Math.PI/60: -2*Math.PI/60;  // -3 => clockwise rotation; -4 => counterclockwise	rotation			
					break;          
				
				case -5:		// rotation from cell center to cell edge => clockwise		
				case -6:		// rotation from cell center to cell edge => counterclockwise												
					// stim = [0 = stim type, 1 = color, 2= polar angle; 3= current radius, 4= radius increase (decrease) in next call]					
					ang = st[j][2];  // units: radians. must always be init in stimGen()
					r = st[j][3];    // units: pixels.  must always be init ok
					x = left + width/2 + r * Math.cos(ang);  
					y = top + height/2 + r * Math.sin(ang);
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();
					// update polar angle in next call depending of stim type						
					st[j][2]+= st[j][0] == -6 ? 2*Math.PI/60: -2*Math.PI/60;  // stim type == 3 => rotation clockwise; else counterclockwise	rotation			
					if ( st[j][3] + 1 > Math.min(width,height)/2 - ball) {
						st[j][4]= -1;  // decrease radius in next call
					}						
					else if ( st[j][3] < 0 ) { 
						st[j][4]= 1;  // increase radius in next call
					}	
					st[j][3]+= st[j][4];						
					break;  
									
				case -7:	 // movement that mimics Archimedean spiral from cell center to cell edge and back => start in cell center with clockwise	rotation
				case -8:   // movement that mimics Archimedean spiral from cell center to cell edge and back => start in cell center with counterclockwise rotation                    						
					// stim = [0 = stim type, 1 = color, 2= polar angle; 3= current radius, 4= radius increase (decrease) in next call]												
					ang = st[j][2];  // units: radians. must always be init in stimGen()
					r = st[j][3];    // units: pixels.  must always be init ok
					x = left + width/2 + r * Math.cos(ang);
					y = top + height/2 + r * Math.sin(ang);
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();   														          			
					// update polar angle in next call depending of stim type	
					st[j][2]+= st[j][0] == -7 ? 2*Math.PI/60: -2*Math.PI/60;  // clockwise or counterclockwiserotation				
					max_r = Math.min(width,height)/2 - ball;
					if ( st[j][3] + 1 >  max_r) {
						st[j][4]= -1;  // decrease radius in next call
					}						
					else if ( st[j][3] - 1 < - max_r ) { 
						st[j][4]= 1;  // increase radius in next call
					}	
					st[j][3]+= st[j][4];   // polar radius in next call									
					break;  
 
				case -9:  // random ball moving within cell walls => clockwise						
					// stim = [0 = stim type, 1 = color, 2 = x, 3 = y, 4 = dx, 5 = dy]
					x= st[j][2], y= st[j][3], dx= st[j][4], dy= st[j][5];
					if (x == null || is_res)  {		 // init ball x,y, start coord somewhere in shape rect (cell) 
            x = getRndInt(left + ball, left + width - ball);
						y = top + ball; 					
					}
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();   
					// check and update dx, dy in next iteration (if ball is out of cell bounds change x and y movement direction) 
					if ( x + dx > left + width - ball ) {  // up-right cell corner (wall) 
						dx= 0;
						dy= ball;
					}
					else if ( y + dy > top + height - ball ) {  // down-right cell corner (wall) 
						dy = 0;
						dx= -ball;
					}
					else if ( x + dx < left + ball ) {  // down-left cell corner 
						dy= -ball;
						dx= 0;
					}				
					else if ( y + dy < top + ball ) {  // up-left cell corner 
						dy= 0;
						dx= ball;
					}															
					st[j][2]= x+dx;
					st[j][3]= y+dy;
					st[j][4]= dx;
					st[j][5]= dy;				
					break;
								
				case -10:  // random ball moving within cell walls => counterclockwise					            
				  // stim = [0 = stim type, 1 = color, 2 = x, 3 = y, 4 = dx, 5 = dy]						
					x= st[j][2], y= st[j][3], dx= st[j][4], dy= st[j][5];
					if (x == null || is_res)  {		 // init ball x,y, start coord somewhere in shape rect (cell) 
						x = getRndInt(left + ball, left + width - ball);
						y = top + ball;          					
					}
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();   
					// check and update dx, dy in next iteration (if ball is out of cell bounds change x and y movement direction) 					
					if ( x + dx < left + ball ) {  // ball come to up-left cell corner 
						dx= 0;
						dy= ball;							
					}					
					else if ( y + dy > top + height - ball ) {  // ball come to down-left cell corner 
						dy= 0;
						dx= ball;
					}
					else if ( x + dx > left + width - ball ) {  // ball come to down-right cell corner 
						dx= 0;
						dy= -ball;
					}				
					else if ( y + dy < top + ball )  {   // ball come to up-right cell corner
						dy= 0;
						dx= -ball;
					}	 				
					st[j][2]= x+dx;
					st[j][3]= y+dy;
					st[j][4]= dx;
					st[j][5]= dy;				
					break;			
					
				default:    // random ball moving within cell walls 					
					// stim = [0 = stim type, 1 = color,  2 = x, 3 = y, 4 = dx, 5 = dy]						
					x= st[j][2], y= st[j][3], dx= st[j][4], dy= st[j][5];
					if (x == null || is_res)  {		 // init ball x,y, start coord somewhere in shape rect (cell) 
						x = getRndInt(left + ball, left + width - ball);
						y = getRndInt(top + ball, top + height - ball);          					
					}
					ctx.beginPath(); 
					ctx.arc(x, y, ball, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
					ctx.fill();   
					// check and update dx, dy in next iteration (if ball is out of cell bounds change x and y movement direction) 
					if ( x + dx < left + ball || x + dx > left + width - ball ) {
						dx = -dx;
					}
					if ( y + dy < top + ball || y + dy > top + height - ball ) {
						dy = -dy;
					}													
					st[j][2]= x+dx;
					st[j][3]= y+dy;
					st[j][4]= dx;
					st[j][5]= dy;				
					//break;
					
			} 	// 	switch ( stype )
			
		}   // else if ( stype <  0 )   
		
	} // for (let j= 0; j < len; j++)	

} 


/* shape function dispatcher (handler)
   ctx = canvas context
   x = x coord rect center,
   y = y coord rect center,
   r = shape radius,   
   shape = shape type => this argument is main for call different shapes.js functions  
   col = color (fillSyle and/or strokeStyle)
   shapes.js file must be included in DOM  (in html file)
	 draws all P-type stimuli     */
function drawShape(ctx, x, y, r, shape, col)
{
  ctx.fillStyle= col;    // fill color
  ctx.strokeStyle= col;  // line color   
  switch (shape)  
  {
    case 0:   // full circle
    case 1:   // 1/2 circle RIGHT oriented
    case 2:   // 1/2 circle DOWN oriented   
    case 3:   // 1/2 circle LEFT oriented
    case 4:   // 1/2 circle UP oriented 
    case 5:   // 3/4 circle with UP-RIGHT hole
    case 6:   // 3/4 circle with DOWN-RIGHT hole
    case 7:   // 3/4 circle with DOWN-LEFT hole 
    case 8:   // 3/4 circle with UP-LEFT hole                          
     if (shape == 0)  {                        
       circle(ctx, x, y, r, 0, 2*Math.PI, false, col);  // full circle 
     }        
     else if (shape == 1)  {
       circle(ctx, x, y, r, 1.5*Math.PI, 0.5*Math.PI, false, col); // 1/2 circle RIGHT oriented      
     }
     else if (shape == 2)  {
       circle(ctx, x, y, r, 0, Math.PI, false, col); // 1/2 circle DOWN oriented     
     } 
     else if (shape == 3)  {
       circle(ctx, x, y, r, 0.5*Math.PI, 1.5*Math.PI, false, col); // 1/2 circle LEFT oriented     
     } 
     else if (shape == 4)  {
       circle(ctx, x, y, r, Math.PI, 0, false, col);   // 1/2 circle UP oriented     
     }        
     else if (shape == 5)  {
       circle(ctx, x, y, r, 0, 1.5*Math.PI, false, col);   // 3/4 circle with UP-RIGHT hole     
     } 
     else if (shape == 6)  { 
       circle(ctx, x, y, r, 0, 0.5*Math.PI, true, col);   // 3/4 circle with DOWN-RIGHT hole     
     } 
     else if (shape == 7)  { 
       circle(ctx, x, y, r, 0.5*Math.PI, Math.PI, true, col);   // 3/4 circle with DOWN-LEFT hole     
     } 
     else /*if (shape == 8)*/  { 
       circle(ctx, x, y, r, 1.5*Math.PI, Math.PI, false, col);   // 3/4 circle with UP-LEFT hole     
     }        
     break;
                
    case 9:   // polygon 3 equal sides
    case 10:  // polygon 4 equal sides   
     (shape == 9) ? polygon(ctx, x, y, r, 3, col):  polygon(ctx, x, y, r, 4, col);       
     break;    
                  
    case 11:  //  xxxxxx
     square(ctx, x, y, r-2, col);  // only for square: window.r must be > 2 for shape selection and demo background fill style  
     break;  //   xxxxxx
             
    case 12:    // longer side (base) down
    case 13:    // longer side (base) up
     (shape == 12) ? trapezoid(ctx, x, y, r, col, true) : trapezoid(ctx, x, y, r, col, false) ;  // base down or up   
     break;
              
    case 14: // rectangle - horizontal line is longer then vertical
    case 15: // rectangle - vertical line is longer then horizontal 
     (shape == 14) ? rectangle(ctx, x, y, r, col, true): rectangle(ctx, x, y, r, col, false);              
     break;
                   
    case 16:
    case 17:
     if (shape == 16) {   // vertical ellipse
      var w= r, h= 2*r; // change this 2 param for PI/2 rotation
     }
     else {   // horizontal ellipse
      var w= 2*r, h= r; // change this 2 param for PI/2 rotation
     }
     ellipse(ctx, x - w / 2.0, y - h / 2.0, w, h, col);       
     break; 
                
    case 18: // cross => classic vertical cross
    case 19: // cross => diagonal cross (x) 
     (shape == 18) ? cross(ctx, x, y, r, col, true): cross(ctx, x, y, r, col, false);       
     break;   
             
    case 20:  //(base == true) => draw down-right rotated paralelogram; 
    case 21: // (base== false) => draw left-down rotated paralelogram
     (shape == 20) ? paralelogram(ctx, x, y, r, col, true): paralelogram(ctx, x, y, r, col, false);    
     break;
                               
    case 22:   // star 4-points
    case 23:   // star 5-points
    case 24:   // star 6-points     
     star(ctx, x, y, shape-18, r/2, r, col);       
     break;
                            
    case 25:  //  vertical UP arrow (0)
    case 26:  //  vertical DOWN arrow (1)
    case 27:  //  horizontal LEFT arrow (2)
    case 28:  //  horizontal RIGHT arrow (3)
     arrow(ctx, x, y, r, col, shape-25);       
     break;
            
    case 29:  
    case 30:
     var a= (shape == 29) ? [r, r, 0, 0] : [0, 0, r, r];    // up rect round radius (31) or down round rect radius (32)
     roundRect(ctx, x-r, y-r, 2*r, 2*r, a, col);    
    break;
       
    case 31:     //  mushroom UP oriented     
    case 32:     //  mushroom RIGHT oriented
    case 33:     //  mushroom DOWN oriented
    case 34:     //  mushroom LEFT orinted 
      mushroom(ctx, x, y, r, col, shape-31);
      break;
               
    case 35:  // polygon 5 equal sides
    case 36:  // polygon 6 equal sides      
      (shape == 35) ?  polygon(ctx, x, y, r, 5, col):  polygon(ctx, x, y, r, 6, col);       
      break; 
                      
  }   // switch (shape) of
   
}  		


// this onclick event is fired by html canvas element => document.getElementById('canvas')
function doAnswer(e)
{ 
  const m_tar= -1;  // target stimulus for M-type
	const p_tar= 9;   // target stimulus for P-type
	const rst = raster;
	let can =  e.target || e.srcElement;  // can = document.getElementById('canvas');      	
	let pos = getCellFromXY(e.offsetX, e.offsetY);

	if ( typeof (pos) === 'undefined' || isNaN(pos) )  {
		return;  // error
	}	
  /*else  {
		pos= parseInt(pos, 10);
	}*/	
  
	// first toogle cell selection
	if ( cells[pos] == 0 ) {
	  cells[pos]= 1;
	}
	else  {
		cells[pos]= 0;  
	}
	let is_sel= cells[pos] == 1 ? true : false; // cell is selected or not
	let p = is_sel ? 1 : -1;	 // number of points (or errors):  increase by 1 or decrease by 1 
	let stype= stim[sec][pos][0];  // stimulus type
	
	if ( stype == m_tar || stype == p_tar )   // is this target stimulus?
	{		
		if (stype == m_tar) { 			
			cor[1]+=p; 	 						
		}	
		else  { 
		  cor[2]+=p; 	 						 
    }	
		// update spatial scores
		let gtype= document.getElementById('gtype').value;
		if ( gtype == '2' )  // spatial left - right
		{		
			if  ( Math.floor (pos % rst) < rst / 2 )  {   
				cor[3]+= p;   // left => index cor[3]
			}
      else if  ( Math.floor (pos % rst) >= rst / 2 )  {
				cor[4]+= p;  // right => index cor[4]
			}				
		} 		
		else if ( gtype == '3' ) // spatial up - down 
		{		 
		  if ( Math.floor (pos / rst) < rst / 2 ) {  
				cor[3]+= p;   // up => index cor[3]
			}	
			else if ( Math.floor (pos / rst) >= rst / 2 ) {
				cor[4]+= p;  // down => index cor[4]
			}			
		}			
	}
	else  {
		errors+= p;	 // increase or decrease error
	} 
	
	drawStim(stim[sec], false);	  // repaint	
}


// end of task (research)
function theEnd()  
{   
  is_task= false;
	
	document.getElementById('canvas').style.display= 'none'; 
  document.getElementById('end').style.display= 'block';   
  document.getElementById('result_wrapper').style.display= 'block'; 
	
	let endTime = new Date();   // see:  https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/getTime
  let task_time= ( endTime.getTime() - startTime.getTime() ) / 1000;  // seconds (convert time difference from milliseconds to seconds)
	task_time= stringTime(task_time);
	 
	let gtype= document.getElementById('gtype').value;
	let header= [], rows = [];  
	let points = cor[1] + cor[2];
	if ( gtype == '1' ) {
	  header= ['   ', 'Total-targets', 'M-targets', 'P-targets'];
		rows.push( [ 'Total',    cor[0], cor[0]/2, cor[0]/2 ] );
		rows.push( [ 'Detected', points, cor[1],   cor[2]   ]);
  }
  else if ( gtype == '2' || gtype == '3' ) {
		header= gtype == '2' ? ['   ', 'Total-targets', 'M-targets', 'P-targets', 'Left', 'Right'] : header= ['   ', 'Total-targets', 'M-targets', 'P-targets', 'Up', 'Down'];
		rows.push( [ 'Total',    cor[0], cor[0]/2, cor[0]/2, cor[0]/2, cor[0]/2 ] );
		rows.push( [ 'Detected', points, cor[1],   cor[2],   cor[3],   cor[4]   ] );
	}
  	     
	createTableArray(document.getElementById('table1'), rows, header);	
	
	document.getElementById('score').textContent= 'Sum of points= ' + points + '. Maximum points= ' + cor[0] + '. Total errors= ' + errors; 
			
	// "chart" => repaint "chart" 
  let ok= Math.round(100 * points / cor[0]).toFixed(2);
	let not_ok= Math.round(100 - ok).toFixed(2); 
	document.getElementById('chart_1').setAttribute('style', 'width:' + ok + '%');    // !important
	document.getElementById('chart_2').setAttribute('style', 'width:' + not_ok + '%');	
	document.getElementById('chart').title= 'The percentage of points scored (green) = ' + ok + '% (100*' + points + '/' + cor[0] + ')';  	
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
		
	try 
  {
    if (is_storage && document.getElementById('store').value === '1') {
			setStorage(cor[0], points, task_time);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }
		
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
	 max_points = maximum possible score in task (number of target stimulus)
	 points = sum of points in task
	 time => task time in hh:mm:ss format   */ 
function setStorage(max_points, points, time)   
{  
  /* dependent variable. if gtype == '1' index: 
   [ 0 => total targets; 1 => M-type targets correct detected; 1 => P-type targets correct detected ] 
   else index: 
   [ 0 => total targets; 1 => M-type targets correct detected; 1 => P-type targets correct detected; 2 => left/up targets correct detected; 3 => right/down targets correct detected] */ 
	let gtype= document.getElementById('gtype').value;
	let oldItems = JSON.parse(localStorage.getItem('pmpt-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(), 					
			'max-points': max_points,
			'points' : points, 	
			'errors' : errors,			
			'm-points': cor[1],
      'p-points': cor[2], 
			'left' : gtype == 2 ? cor[3] : '', 
			'right': gtype == 2 ? cor[4] : '', 
			'up' :   gtype == 3 ? cor[3] : '', 
			'down' : gtype == 3 ? cor[4] : '', 			
			'demo' : document.getElementById('demo').checked ? 1 : 0,
			'minute': minute,
			'generator' : gtype,
			'raster': raster,
			'color' : document.getElementById('tcolor').value,
			'total-time': time
  };

  oldItems.push(newItem);

  localStorage.setItem('pmpt-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
   
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


/* Get the cell (rectangle) ordinal position in html canvas element from x and y mouse position
   offsetX:  x - coordinate of the mouse pointer, relative to the target element. 
   offsetY:	 y - coordinate of the mouse pointer, relative to the target element. */
function getCellFromXY(offsetX, offsetY)
{	
  let canvas= document.getElementById('canvas');
	let dx =  canvas.width / raster, dy = canvas.height / raster;  
  let row = Math.floor ( offsetY / dy );  // The offsetX (offsetY) property returns the x (y) -coordinate of the mouse pointer, relative to the target element. 
  let col = Math.floor ( offsetX / dx );  // The clientX (clientY) property returns the horizontal (vertical) coordinate (according to the client area) of the mouse pointer when a mouse event was triggered.
  if (row > raster - 1 || row < 0 || col > raster - 1 || col < 0) {
    return;  // return 'undefined' value	=> abort all     
  }
  let pos = row * raster + col;   // rectangle position within html canvas object  	
	if ( pos >= raster * raster || pos < 0 ) {    
		return; // return 'undefined' value	=> abort all  
  }
	else {		
		return pos;	
	}
}	

 
// begin of new code for export localStorage app data:  data= localStorage.getItem('myApp-v1.0');

 
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').setAttribute('class', 'center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (document.getElementById('store_group').disabled === false && localStorage.getItem('pmpt-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('pmpt-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('pmpt-v1.0') ? JSON.parse(localStorage.getItem('pmpt-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application data:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('pmpt-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('pmpt-v1.0');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('pmpt-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object id (div or table)
function newWinLast(id)
{ 
	let e = document.getElementById('table1');		
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{     
  let tbl= document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element		
	if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
